/* 
Atheeshan VIGNESWARAN, Ismaïl ABLOUA
Group 7
*/
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>
#include <unistd.h>

#define NO_WAIT
#ifdef NO_WAIT
#define uspleep(x) ((void)0)
#else
#define uspleep(x) usleep(x)
#endif

// structure site
typedef struct {
    int available_bikes;
    int wait_put;
    int wait_get;
    sem_t sem_bikes;
    sem_t sem_slots;
    pthread_mutex_t site_mutex;
} site_t;

// global constants
#define MAX_BIKES_TRUCK 4

// global variables
int S = 0; // nb sites
int H = 0; // nb habitants
int M = 0; // nb travel by habitants
int N = 0; // nb slots per site

site_t* sites;

pthread_mutex_t mutex_finished_habitants;
int finished_habitants = 0;

int truck = 2;
int depot = 0;

void habitant_routine(void* arg) {
    int id_habitant = *(int*)arg;
    int current_site = rand() % S;

    for(int i = 0; i < M; i++) {

        // init wait get
        pthread_mutex_lock(&sites[current_site].site_mutex);
        sites[current_site].wait_get++;
        pthread_mutex_unlock(&sites[current_site].site_mutex);

        // get bike from current site
        sem_wait(&sites[current_site].sem_bikes);
        pthread_mutex_lock(&sites[current_site].site_mutex);
        sites[current_site].available_bikes--;
        sites[current_site].wait_get--;
        printf("(GET) Habitant %d taking bike from site %d (%d bikes remaining, %d people waiting)\n", id_habitant, current_site, sites[current_site].available_bikes, sites[current_site].wait_get);
        sem_post(&sites[current_site].sem_slots);
        pthread_mutex_unlock(&sites[current_site].site_mutex);

        // Travel time
        usleep(1000 + rand() % 1000);
        // travel to destination site
        int destination_site;
        do {destination_site = rand() % S;}
        while (destination_site == current_site);
        current_site = destination_site;

        // init wait put
        pthread_mutex_lock(&sites[current_site].site_mutex);
        sites[current_site].wait_put++;
        pthread_mutex_unlock(&sites[current_site].site_mutex);
        
        // return bike to destination site
        sem_wait(&sites[current_site].sem_slots);
        pthread_mutex_lock(&sites[current_site].site_mutex);
        sites[current_site].available_bikes++;
        sites[current_site].wait_put--;
        printf("(PUT) Habitant %d returning bike to site %d (%d bikes remaining, %d people waiting)\n", id_habitant, current_site, sites[current_site].available_bikes, sites[current_site].wait_put);
        sem_post(&sites[current_site].sem_bikes);
        pthread_mutex_unlock(&sites[current_site].site_mutex);

        // Activity time
        usleep(1000 + rand() % 1000);
    }

    // finished
    printf("Habitant %d finished\n", id_habitant);
    pthread_mutex_lock(&mutex_finished_habitants);
    finished_habitants++;
    pthread_mutex_unlock(&mutex_finished_habitants);
}

void truck_routine() {
    // check if all habitants are finished
    while (1) {
        pthread_mutex_lock(&mutex_finished_habitants);
        if (finished_habitants == H) {
            pthread_mutex_unlock(&mutex_finished_habitants);
            printf("(TRUCK) Truck leaving as all habitants have finished their routine.\n");
            break;
        }
        pthread_mutex_unlock(&mutex_finished_habitants);

        // refill site routine
        for(int i = 0; i < S; i++){
            pthread_mutex_lock(&sites[i].site_mutex);
            while(truck > 0 && sites[i].available_bikes < 2) {
                sem_wait(&sites[i].sem_slots);
                truck--;
                sites[i].available_bikes++;
                sem_post(&sites[i].sem_bikes);
                printf("(TRUCK) Truck refilling site %d (%d bikes remaining)\n", i, sites[i].available_bikes);
            }
            while(truck < MAX_BIKES_TRUCK && sites[i].available_bikes > N-2) {
                sem_wait(&sites[i].sem_bikes);
                sites[i].available_bikes--;
                truck++;
                sem_post(&sites[i].sem_slots);
                printf("(TRUCK) Truck taking bike from site %d (%d bikes remaining)\n", i, sites[i].available_bikes);
            }
            pthread_mutex_unlock(&sites[i].site_mutex);
        }

        // refill depot routine
        while(truck > 2) {
            printf("(DEPOT) Truck returning to depot (%d bikes in truck)\n", truck);
            truck--;
            depot++;
        }
        while(truck < 2 && depot > 0) {
            printf("(DEPOT) Truck returning to depot (%d bikes in truck)\n", truck);
            depot--;
            truck++;
        }

        usleep(100 + rand() % 100);
    }
}

int main(int argc, char* argv[]) {
    if (argc != 5) {
        printf("Usage: %s <S> <H> <M> <N>\n", argv[0]);
        return EXIT_FAILURE;
    }

    // Read arguments
    S = atoi(argv[1]);
    H = atoi(argv[2]);
    M = atoi(argv[3]);
    N = atoi(argv[4]);

    if(S <= 2 || H <= 2 || M <= 2 || N <= 2) {
        perror("Must be: S > 2, H > 2, M > 2, N > 2");
        return EXIT_FAILURE;
    }

    // init random
    srand(time(NULL));

    // init mutex
    pthread_mutex_init(&mutex_finished_habitants, NULL);
    finished_habitants = 0;

    // init sites
    sites = malloc(S * sizeof(*sites));
    for (int i = 0; i < S; i++) {
        sites[i].available_bikes = N-2;
        sem_init(&sites[i].sem_bikes, 0, N-2);
        sem_init(&sites[i].sem_slots, 0, 2);
        sites[i].wait_get = sites[i].wait_put = 0;
        pthread_mutex_init(&sites[i].site_mutex, NULL);
    }

    // init habitants
    int* id_habitants = malloc(H * sizeof(*id_habitants));
    pthread_t* threads_habitants = malloc(H * sizeof(*threads_habitants));
    for(int i = 0; i < H; i++) {
        id_habitants[i] = i;
        pthread_create(&threads_habitants[i], NULL, (void*)habitant_routine, &id_habitants[i]);
    }
    // init truck
    truck = 2;
    depot = 0;
    pthread_t thread_truck;
    pthread_create(&thread_truck, NULL, (void*)truck_routine, NULL);

    // wait for threads to finish
    for(int i = 0; i < H; i++) {
        pthread_join(threads_habitants[i], NULL);
    }
    pthread_join(thread_truck, NULL);

    //print status; all sites bikes, truck bikes, depot bikes
    printf("Final status:\n");
    for(int i = 0; i < S; i++) {
        printf("Site %d: %d bikes\n", i, sites[i].available_bikes);
    }
    printf("Truck: %d bikes\n", truck);
    printf("Depot: %d bikes\n", depot);

    //simple math to check if all bikes are accounted for
    int total_bikes = S*(N-2) + 2;
    int calculated_bikes = 0;
    for(int i = 0; i < S; i++) {
        calculated_bikes += sites[i].available_bikes;
    }
    calculated_bikes += truck;
    calculated_bikes += depot;
    printf("Total bikes: %d\n", total_bikes);
    if(calculated_bikes != total_bikes) {
        printf("Error: calculated bikes %d != total bikes %d\n", calculated_bikes, total_bikes);
    } else {
        printf("All bikes accounted for.\n");
    }

    // free memory
    for (int i = 0; i < S; i++) {
        sem_destroy(&sites[i].sem_bikes);
        sem_destroy(&sites[i].sem_slots);
        pthread_mutex_destroy(&sites[i].site_mutex);
    }
    free(sites);
    free(id_habitants);
    free(threads_habitants);
    pthread_mutex_destroy(&mutex_finished_habitants);

    return EXIT_SUCCESS;
}
